#region licence/info

//////project name
//vvvv plugin template

//////description
//basic vvvv node plugin template.
//Copy this an rename it, to write your own plugin node.

//////licence
//GNU Lesser General Public License (LGPL)
//english: http://www.gnu.org/licenses/lgpl.html
//german: http://www.gnu.de/lgpl-ger.html

//////language/ide
//C# sharpdevelop 

//////dependencies
//VVVV.PluginInterfaces.V1;
//VVVV.Utils.VColor;
//VVVV.Utils.VMath;

//////initial author
//vvvv group

#endregion licence/info

//use what you need
using System;
using System.Drawing;
using VVVV.PluginInterfaces.V1;
using VVVV.Utils.VColor;
using VVVV.Utils.VMath;

//the vvvv node namespace
namespace VVVV.Nodes
{
	
	//class definition
	public class PolyVal2D: IPlugin, IDisposable
    {	          	
    	#region field declaration
    	
    	//the host (mandatory)
    	private IPluginHost FHost; 
    	// Track whether Dispose has been called.
   		private bool FDisposed = false;

    	//input pin declaration
        private IValueIn FWidthIn, FHeightIn;
        private IValueIn FInterleaveWidthIn, FInterleaveHeightIn;
        private IValueIn FNParityBits;

        private IValueIn FFrameIn;
    	
    	//output pin declaration
    	private IValueOut FiInterleaveOut, FInterleaveFrameOut, FNCalibrationFramesOut;
        private IValueOut FNFramesOut, FNFramesPerInterleaveOut, FNInterleavesOut;
        private IValueOut FNisCalibrating, FNisCRCOut;
    	
    	#endregion field declaration
       
    	#region constructor/destructor
    	
        public PolyVal2D()
        {
			//the nodes constructor
			//nothing to declare for this node
		}
        
        // Implementing IDisposable's Dispose method.
        // Do not make this method virtual.
        // A derived class should not be able to override this method.
        public void Dispose()
        {
        	Dispose(true);
        	// Take yourself off the Finalization queue
        	// to prevent finalization code for this object
        	// from executing a second time.
        	GC.SuppressFinalize(this);
        }
        
        // Dispose(bool disposing) executes in two distinct scenarios.
        // If disposing equals true, the method has been called directly
        // or indirectly by a user's code. Managed and unmanaged resources
        // can be disposed.
        // If disposing equals false, the method has been called by the
        // runtime from inside the finalizer and you should not reference
        // other objects. Only unmanaged resources can be disposed.
        protected virtual void Dispose(bool disposing)
        {
        	// Check to see if Dispose has already been called.
        	if(!FDisposed)
        	{
        		if(disposing)
        		{
        			// Dispose managed resources.
        		}
        		// Release unmanaged resources. If disposing is false,
        		// only the following code is executed.
	        	
        		FHost.Log(TLogType.Debug, "PluginTemplate is being deleted");
        		
        		// Note that this is not thread safe.
        		// Another thread could start disposing the object
        		// after the managed resources are disposed,
        		// but before the disposed flag is set to true.
        		// If thread safety is necessary, it must be
        		// implemented by the client.
        	}
        	FDisposed = true;
        }

        // Use C# destructor syntax for finalization code.
        // This destructor will run only if the Dispose method
        // does not get called.
        // It gives your base class the opportunity to finalize.
        // Do not provide destructors in types derived from this class.
        ~PolyVal2D()
        {
        	// Do not re-create Dispose clean-up code here.
        	// Calling Dispose(false) is optimal in terms of
        	// readability and maintainability.
        	Dispose(false);
        }
        #endregion constructor/destructor
        
        #region node name and infos
       
        //provide node infos 
        private static IPluginInfo FPluginInfo;
        public static IPluginInfo PluginInfo
	    {
	        get 
	        {
	        	if (FPluginInfo == null)
				{
					//fill out nodes info
					//see: http://www.vvvv.org/tiki-index.php?page=Conventions.NodeAndPinNaming
					FPluginInfo = new PluginInfo();
					
					//the nodes main name: use CamelCaps and no spaces
					FPluginInfo.Name = "PCEncode calculator";
					//the nodes category: try to use an existing one
					FPluginInfo.Category = "Value";
					//the nodes version: optional. leave blank if not
					//needed to distinguish two nodes of the same name and category
					FPluginInfo.Version = "";
					
					//the nodes author: your sign
					FPluginInfo.Author = "sugokuGENKI";
					//describe the nodes function
					FPluginInfo.Help = "Evaluates variables for PCEncode";
					//specify a comma separated list of tags that describe the node
					FPluginInfo.Tags = "";
					
					//give credits to thirdparty code used
					FPluginInfo.Credits = "";
					//any known problems?
					FPluginInfo.Bugs = "";
					//any known usage of the node that may cause troubles?
					FPluginInfo.Warnings = "";
					
					//leave below as is
					System.Diagnostics.StackTrace st = new System.Diagnostics.StackTrace(true);
					System.Diagnostics.StackFrame sf = st.GetFrame(0);
					System.Reflection.MethodBase method = sf.GetMethod();
					FPluginInfo.Namespace = method.DeclaringType.Namespace;
					FPluginInfo.Class = method.DeclaringType.Name;
					//leave above as is
				}
				return FPluginInfo;
	        }
		}

        public bool AutoEvaluate
        {
        	//return true if this node needs to calculate every frame even if nobody asks for its output
        	get {return false;}
        }
        
        #endregion node name and infos
        
      	#region pin creation
        
        //this method is called by vvvv when the node is created
        public void SetPluginHost(IPluginHost Host)
	    {
        	//assign host
	    	FHost = Host;

            //create inputs
            FHost.CreateValueInput("Width", 1, null, TSliceMode.Single, TPinVisibility.True, out FWidthIn);
            FWidthIn.SetSubType(1, double.MaxValue, 1, 320, false, false, true);

            FHost.CreateValueInput("Height", 1, null, TSliceMode.Single, TPinVisibility.True, out FHeightIn);
            FHeightIn.SetSubType(1, double.MaxValue, 1, 240, false, false, true);

            FHost.CreateValueInput("InterleaveWidth", 1, null, TSliceMode.Single, TPinVisibility.True, out FInterleaveWidthIn);
            FInterleaveWidthIn.SetSubType(1, double.MaxValue, 1, 1, false, false, true);
            FHost.CreateValueInput("InterleaveHeight", 1, null, TSliceMode.Single, TPinVisibility.True, out FInterleaveHeightIn);
            FInterleaveHeightIn.SetSubType(1, double.MaxValue, 1, 1, false, false, true);

            FHost.CreateValueInput("NParityBits", 1, null, TSliceMode.Single, TPinVisibility.True, out FNParityBits);
            FNParityBits.SetSubType(1, double.MaxValue, 1, 16, false, false, true);

            FHost.CreateValueInput("Frame", 1, null, TSliceMode.Single, TPinVisibility.True, out FFrameIn);
            FFrameIn.SetSubType(0, double.MaxValue, 1, 0, false, false, true);
	    	
	    	//create outputs	    	
            FHost.CreateValueOutput("NFrames", 1, null, TSliceMode.Single, TPinVisibility.True, out FNFramesOut);
            FNFramesOut.SetSubType(double.MinValue, double.MaxValue, 1, 0, false, false, true);

            FHost.CreateValueOutput("NFramesPerInterleave", 1, null, TSliceMode.Single, TPinVisibility.True, out FNFramesPerInterleaveOut);
            FNFramesPerInterleaveOut.SetSubType(double.MinValue, double.MaxValue, 1, 0, false, false, true);

            FHost.CreateValueOutput("NInterleaves", 1, null, TSliceMode.Single, TPinVisibility.True, out FNInterleavesOut);
            FNInterleavesOut.SetSubType(double.MinValue, double.MaxValue, 1, 0, false, false, true);

            FHost.CreateValueOutput("NCalibrationFrames", 1, null, TSliceMode.Single, TPinVisibility.True, out FNCalibrationFramesOut);
            FNCalibrationFramesOut.SetSubType(0, double.MaxValue, 1, 0, false, false, true);

            FHost.CreateValueOutput("iInterleave", 1, null, TSliceMode.Single, TPinVisibility.True, out FiInterleaveOut);
            FiInterleaveOut.SetSubType(double.MinValue, double.MaxValue, 1, 0, false, false, true);

            FHost.CreateValueOutput("iInterleaveFrame", 1, null, TSliceMode.Single, TPinVisibility.True, out FInterleaveFrameOut);
            FInterleaveFrameOut.SetSubType(double.MinValue, double.MaxValue, 1, 0, false, false, true);

            FHost.CreateValueOutput("isCalibrating", 1, null, TSliceMode.Single, TPinVisibility.True, out FNisCalibrating);
            FNisCalibrating.SetSubType(0, 1, 1, 0, false, false, true);

            FHost.CreateValueOutput("isCRC", 1, null, TSliceMode.Single, TPinVisibility.True, out FNisCRCOut);
            FNisCRCOut.SetSubType(0, 1, 1, 0, false, false, true);
        }

        #endregion pin creation
        
        #region mainloop
        
        public void Configurate(IPluginConfig Input)
        {
        	//nothing to configure in this plugin
        	//only used in conjunction with inputs of type cmpdConfigurate
        }
        
        //here we go, thats the method called by vvvv each frame
        //all data handling should be in here
        public void Evaluate(int SpreadMax)
        {     	
        	//if any of the inputs has changed
        	//recompute the outputs
            if (FWidthIn.PinIsChanged || FHeightIn.PinIsChanged || FInterleaveWidthIn.PinIsChanged || FInterleaveHeightIn.PinIsChanged || FFrameIn.PinIsChanged)
            {
                //set values used across all slices
                double Width, Height, InterleaveWidth, InterleaveHeight, iFrame, nParityBits;

                //inputs
                FWidthIn.GetValue(0, out Width);
                FHeightIn.GetValue(0, out Height);
                FInterleaveWidthIn.GetValue(0, out InterleaveWidth);
                FInterleaveHeightIn.GetValue(0, out InterleaveHeight);
                FNParityBits.GetValue(0, out nParityBits); 
                FFrameIn.GetValue(0, out iFrame);

                //calcs
                double nFramesPerInterleave = Math.Ceiling(Math.Log((Width*Height)/(InterleaveWidth*InterleaveHeight),2))+nParityBits;
                double nInterleaves = InterleaveWidth * InterleaveHeight;
                double nCalibrationFrames = nInterleaves + 1;
                double nFrames = nFramesPerInterleave * nInterleaves + nCalibrationFrames;

                iFrame -= nCalibrationFrames;

                double iInterleave = Math.Floor(iFrame / nFramesPerInterleave);
                double iInterleaveFrame = iFrame % nFramesPerInterleave;

                bool isCalibrating = iFrame < 0;
                bool isCRC = iInterleaveFrame >= nFramesPerInterleave - nParityBits;

                if (isCalibrating)
                {
                    iInterleave = 0;
                    iInterleaveFrame = 0;
                }

                //outputs
                FNFramesPerInterleaveOut.SetValue(0, nFramesPerInterleave);
                FNFramesOut.SetValue(0, nFrames);
                FNCalibrationFramesOut.SetValue(0, nCalibrationFrames);
                FiInterleaveOut.SetValue(0, iInterleave);
                FInterleaveFrameOut.SetValue(0, iInterleaveFrame);
                FNInterleavesOut.SetValue(0, nInterleaves);
                FNisCalibrating.SetValue(0, (isCalibrating ? 1 : 0));
                FNisCRCOut.SetValue(0, (isCRC ? 1 : 0));
        	}      	
        }
             
        #endregion mainloop  
	}
}
