﻿using System;
using System.Collections.Generic;
using System.Text;
using VVVV.Utils.VMath;

namespace VVVV.Nodes
{
    class SMTSlide : IComparable<SMTSlide>
    {
        private Vector2D _Pos, _Sca, _Vel;
        private double _Rot;
        private int _Slide, _LastCursorID, _ThisCursorID;
        private int _CursorUniqueID;

        private int _ID, _UniqueID; //unique id is different for every slide ever
                                    //id denotes the image being used

        private Vector2D _deltaxy;
        private Vector2D _primarycursorxy;
        private Vector2D _cursorstartMTxy1;
        private Vector2D _cursorstartMTxy2;

        private double _deltascale, _deltatheta;

        private DateTime _LastTouched;

        private Matrix4x4 _Matrix;

        private bool _isMoving, _isMT;

        public SMTSlide(Vector2D inPos, double inRot, Vector2D inSca, Vector2D inVel, int inSlide, int inLastCursorID, int inID, int inUniqueID)
        {
            _Pos = inPos;
            _Rot = inRot;
            _Sca = inSca;
            _Vel = inVel;
            _Slide = inSlide;
            _LastCursorID = inLastCursorID;
            _ThisCursorID = -1;
            _ID = inID;
            _UniqueID = inUniqueID;
            _LastTouched = DateTime.Now;
            fnGenerateMatrix();
            _isMoving = false;
            _isMT = false;
            _deltascale = 1;
        }

        void fnGenerateMatrix()
        {
            _Matrix=VMath.Scale(new Vector3D(_Sca.x,_Sca.y,1));
            _Matrix *= VMath.RotateZ(VMath.CycToRad*_Rot);
            _Matrix *= VMath.Translate(new Vector3D(_Pos));
        }

        public void doTouch()
        {
            _LastTouched = DateTime.Now;
        }

        public Matrix4x4 getMatrixMT()
        {
            Matrix4x4 MatOut = VMath.IdentityMatrix;
            Matrix4x4 MatPrimaryFingerSpace = VMath.Translate(_cursorstartMTxy1.x, _cursorstartMTxy1.y, 0);

            MatOut *= !MatPrimaryFingerSpace;
            MatOut *= VMath.RotateZ(_deltatheta);
            MatOut *= VMath.Scale(_deltascale, _deltascale, 1);
            MatOut *= MatPrimaryFingerSpace;
            return MatOut;
        }

        public Matrix4x4 getMatrix()
        {
            if (_isMoving)
            {
                Matrix4x4 MatOut;
                
                MatOut = _Matrix;

                MatOut *= getMatrixMT();

                MatOut *= VMath.Translate(_deltaxy.x, _deltaxy.y, 0);
                return MatOut;
            }
            else
                return _Matrix;
        }

        public void bakeMTMat()
        {
            _Matrix *= getMatrixMT();
            _deltascale = 1;
            _deltatheta = 0;
        }


        public void setdeltaxy(Vector2D indeltaxy)
        {
            _deltaxy=indeltaxy;
            doTouch();
        }

        public void setdeltascale(double inscale)
        {
            _deltascale = inscale;
        }
        public void setdeltarot(double theta)
        {
            _deltatheta = theta;
        }

        public bool hasCursor()
        {
            return !(_ThisCursorID < 0);
        }

        public double LastCursorID
        {
            get
            {
                return System.Convert.ToDouble(_Slide);
            }
            set
            {
                _LastCursorID = System.Convert.ToInt32(value);
            }
        }

        public DateTime LastTouched
        {
            get
            {
                return _LastTouched;
            }

            set
            {
                _LastTouched = value;
            }
        }

        public int ID
        {
            get
            {
                return _ID;
            }

            set
            {
                _ID = value;
            }
        }

        public int UniqueID
        {
            get
            {
                return _UniqueID;
            }
            set
            {
                _UniqueID = value;
            }
        }

        public int CompareTo(SMTSlide other)
        {
            return other.LastTouched.CompareTo(_LastTouched);
        }

        public void clearCursor()
        {
            _ThisCursorID = -1;
        }

        public bool isMoving
        {
            get
            {
                return _isMoving;
            }
            set
            {
                if (value == false)
                {
                    _Matrix = getMatrix();
                    _deltascale = 1;
                }

                _isMoving = value;
            }
        }

        public bool isMT
        {
            get
            {
                return _isMT;
            }
            set
            {
                if (value == false)
                {
                    bakeMTMat();
                }
                else
                {
                    _deltatheta = 0;
                    _deltascale = 0;
                }


                _isMT = value;
            }
        }

        public Vector2D primarycursorxy
        {
            get
            {
                return _primarycursorxy;
            }

            set
            {
                _primarycursorxy = value;
            }
        }

        public Vector2D cursorstartMTxy1
        {
            get
            {
                return _cursorstartMTxy1;
            }

            set
            {
                _cursorstartMTxy1 = value;
            }
        }

        public Vector2D cursorstartMTxy2
        {
            get
            {
                return _cursorstartMTxy2;
            }

            set
            {
                _cursorstartMTxy2 = value;
            }
        }

        public int cursoruniqueid
        {
            get
            {
                return _CursorUniqueID;
            }
            set
            {
                _CursorUniqueID = value;
            }
        }

    }
}
