#region licence/info

//////project name
//vvvv plugin template

//////description
//basic vvvv node plugin template.
//Copy this an rename it, to write your own plugin node.

//////licence
//GNU Lesser General Public License (LGPL)
//english: http://www.gnu.org/licenses/lgpl.html
//german: http://www.gnu.de/lgpl-ger.html

//////language/ide
//C# sharpdevelop 

//////dependencies
//VVVV.PluginInterfaces.V1;
//VVVV.Utils.VColor;
//VVVV.Utils.VMath;

//////initial author
//vvvv group

#endregion licence/info

//use what you need
using System;
using System.Drawing;
using System.Collections.Generic;
using VVVV.PluginInterfaces.V1;
using VVVV.Utils.VColor;
using VVVV.Utils.VMath;

//the vvvv node namespace
namespace VVVV.Nodes
{
	
	//class definition
	public class stack_mt: IPlugin, IDisposable
    {	          	
    	#region field declaration
    	
    	//the host (mandatory)
    	private IPluginHost FHost; 
    	// Track whether Dispose has been called.
   		private bool FDisposed = false;

    	//input pin declaration
        private IValueIn PinInFingerXY, PinInFingerID;
        private IValueIn PinInSizeCloseButton;
        private IValueIn PinInAddPosXY, PinInAddRot, PinInAddScaXY, PinInAddVelXY;
        private IValueIn PinInAddIndex, PinInDoAdd, PinInReset;

    	//output pin declarations
    	private ITransformOut PinOutTransform;
    	private IValueOut PinOutIndex, PinOutCount;
        private IValueOut PinOutSlideCursor;
        private IValueOut PinOutOver;

        //storage variables
        private List<SMTSlide> _SlideList = new List<SMTSlide>();
        List<SMTCursor> _CursorList = new List<SMTCursor>();
        int UniqueSlideID = 0, UniqueCursorID = 0;

        //current variables
        private List<SMTIntersect> _IntersectList = new List<SMTIntersect>();

    	#endregion field declaration
       
    	#region constructor/destructor
    	
        public stack_mt()
        {
			//the nodes constructor
			//nothing to declare for this node
		}
        
        // Implementing IDisposable's Dispose method.
        // Do not make this method virtual.
        // A derived class should not be able to override this method.
        public void Dispose()
        {
        	Dispose(true);
        	// Take yourself off the Finalization queue
        	// to prevent finalization code for this object
        	// from executing a second time.
        	GC.SuppressFinalize(this);
        }
        
        // Dispose(bool disposing) executes in two distinct scenarios.
        // If disposing equals true, the method has been called directly
        // or indirectly by a user's code. Managed and unmanaged resources
        // can be disposed.
        // If disposing equals false, the method has been called by the
        // runtime from inside the finalizer and you should not reference
        // other objects. Only unmanaged resources can be disposed.
        protected virtual void Dispose(bool disposing)
        {
        	// Check to see if Dispose has already been called.
        	if(!FDisposed)
        	{
        		if(disposing)
        		{
        			// Dispose managed resources.
        		}
        		// Release unmanaged resources. If disposing is false,
        		// only the following code is executed.
	        	
        		FHost.Log(TLogType.Debug, "PluginTemplate is being deleted");
        		
        		// Note that this is not thread safe.
        		// Another thread could start disposing the object
        		// after the managed resources are disposed,
        		// but before the disposed flag is set to true.
        		// If thread safety is necessary, it must be
        		// implemented by the client.
        	}
        	FDisposed = true;
        }

        // Use C# destructor syntax for finalization code.
        // This destructor will run only if the Dispose method
        // does not get called.
        // It gives your base class the opportunity to finalize.
        // Do not provide destructors in types derived from this class.
        ~stack_mt()
        {
        	// Do not re-create Dispose clean-up code here.
        	// Calling Dispose(false) is optimal in terms of
        	// readability and maintainability.
        	Dispose(false);
        }
        #endregion constructor/destructor
        
        #region node name and infos
       
        //provide node infos 
        private static IPluginInfo FPluginInfo;
        public static IPluginInfo PluginInfo
	    {
	        get 
	        {
	        	if (FPluginInfo == null)
				{
					//fill out nodes info
					//see: http://www.vvvv.org/tiki-index.php?page=Conventions.NodeAndPinNaming
					FPluginInfo = new PluginInfo();
					
					//the nodes main name: use CamelCaps and no spaces
					FPluginInfo.Name = "Stack";
					//the nodes category: try to use an existing one
					FPluginInfo.Category = "MultiTouch";
					//the nodes version: optional. leave blank if not
					//needed to distinguish two nodes of the same name and category
					FPluginInfo.Version = "";
					
					//the nodes author: your sign
					FPluginInfo.Author = "sugokuGENKI";
					//describe the nodes function
					FPluginInfo.Help = "Gives positions for items in a multitouch stack";
					//specify a comma separated list of tags that describe the node
					FPluginInfo.Tags = "";
					
					//give credits to thirdparty code used
					FPluginInfo.Credits = "";
					//any known problems?
					FPluginInfo.Bugs = "";
					//any known usage of the node that may cause troubles?
					FPluginInfo.Warnings = "";
					
					//leave below as is
					System.Diagnostics.StackTrace st = new System.Diagnostics.StackTrace(true);
					System.Diagnostics.StackFrame sf = st.GetFrame(0);
					System.Reflection.MethodBase method = sf.GetMethod();
					FPluginInfo.Namespace = method.DeclaringType.Namespace;
					FPluginInfo.Class = method.DeclaringType.Name;
					//leave above as is
				}
				return FPluginInfo;
	        }
		}

        public bool AutoEvaluate
        {
        	//return true if this node needs to calculate every frame even if nobody asks for its output
        	get {return false;}
        }
        
        #endregion node name and infos
        
      	#region pin creation
        
        //this method is called by vvvv when the node is created
        public void SetPluginHost(IPluginHost Host)
	    {
        	//assign host
	    	FHost = Host;

            //create inputs
            FHost.CreateValueInput("Finger", 2, null, TSliceMode.Dynamic, TPinVisibility.True, out PinInFingerXY);
            PinInFingerXY.SetSubType2D(double.MinValue, double.MaxValue, 0.001, 0, 0, false, false, false);

            FHost.CreateValueInput("Finger ID", 1, null, TSliceMode.Dynamic, TPinVisibility.True, out PinInFingerID);
            PinInFingerID.SetSubType(double.MinValue, double.MaxValue, 0.001, 0, false, false, true);

            FHost.CreateValueInput("Close button size", 2, null, TSliceMode.Single, TPinVisibility.True, out PinInSizeCloseButton);
            PinInSizeCloseButton.SetSubType2D(double.MinValue, double.MaxValue, 0.001, 0,0, false, false, false);

            FHost.CreateValueInput("Add position", 2, null, TSliceMode.Dynamic, TPinVisibility.True, out PinInAddPosXY);
            PinInAddPosXY.SetSubType2D(double.MinValue, double.MaxValue, 0.001, 0, 0, false, false, false);
            FHost.CreateValueInput("Add rotation", 1, null, TSliceMode.Dynamic, TPinVisibility.True, out PinInAddRot);
            PinInAddRot.SetSubType(double.MinValue, double.MaxValue, 0.001, 0, false, false, false);
            FHost.CreateValueInput("Add scale", 2, null, TSliceMode.Dynamic, TPinVisibility.True, out PinInAddScaXY);
            PinInAddScaXY.SetSubType2D(double.MinValue, double.MaxValue, 0.001, 0, 0, false, false, false);
            FHost.CreateValueInput("Add velocity", 2, null, TSliceMode.Dynamic, TPinVisibility.True, out PinInAddVelXY);
            PinInAddVelXY.SetSubType2D(double.MinValue, double.MaxValue, 0.001, 0, 0, false, false, false);

            FHost.CreateValueInput("Add index", 1, null, TSliceMode.Dynamic, TPinVisibility.True, out PinInAddIndex);
            PinInAddIndex.SetSubType(0, double.MaxValue, 1, 0, false, false, true);

            FHost.CreateValueInput("Add", 1, null, TSliceMode.Dynamic, TPinVisibility.True, out PinInDoAdd);
            PinInDoAdd.SetSubType(0, 1, 1, 0, true, false, true);

            FHost.CreateValueInput("Reset", 1, null, TSliceMode.Dynamic, TPinVisibility.True, out PinInReset);
            PinInReset.SetSubType(0, 1, 1, 0, true, false, true);

	    	
	    	//create outputs
	    	
            FHost.CreateTransformOutput("Transform", TSliceMode.Dynamic, TPinVisibility.True, out PinOutTransform);

            FHost.CreateValueOutput("Index", 1, null, TSliceMode.Dynamic, TPinVisibility.True, out PinOutIndex);
            PinOutIndex.SetSubType(0, double.MaxValue, 1, 0, false, false, true);

            FHost.CreateValueOutput("Count", 1, null, TSliceMode.Single, TPinVisibility.True, out PinOutCount);
            PinOutCount.SetSubType(0, double.MaxValue, 1, 0, false, false, true);

            FHost.CreateValueOutput("Slide Cursor", 2, null, TSliceMode.Dynamic, TPinVisibility.True, out PinOutSlideCursor);
            PinOutSlideCursor.SetSubType(double.MinValue, double.MaxValue, 0.001, 0, false, false, false);
            
            FHost.CreateValueOutput("Selected", 1, null, TSliceMode.Dynamic, TPinVisibility.True, out PinOutOver);
            PinOutOver.SetSubType(0, 1, 1, 0, false, true, true);
        }

        #endregion pin creation
        
        #region mainloop
        
        public void Configurate(IPluginConfig Input)
        {
        	//nothing to configure in this plugin
        	//only used in conjunction with inputs of type cmpdConfigurate
        }
        
        public void Evaluate(int SpreadMax)
        {
     	    double currentAdd;
            double currentReset;

            PinInDoAdd.GetValue(0, out currentAdd);
            PinInReset.GetValue(0, out currentReset);

            if (currentAdd > 0)
                fnAddSlide();

            if (currentReset > 0)
                fnReset();

            fnFindCursors();

            fnFindIntersects();

            fnMaybeClose();

            fnMoveSlides();

            fnOutput();

        }
        void fnAddSlide()
        {
            Vector2D inPos, inSca, inVel;
            double inRot, inSlide;

            int count = PinInAddIndex.SliceCount;
            if ((PinInAddIndex.SliceCount > 0) && (PinInAddPosXY.SliceCount > 0) && (PinInAddRot.SliceCount > 0) && (PinInAddScaXY.SliceCount > 0) && (PinInAddVelXY.SliceCount>0))            
            {
                for (int i = 0; i<count; i++)
                {
                    PinInAddPosXY.GetValue2D(i, out inPos.x, out inPos.y);
                    PinInAddRot.GetValue(i, out inRot);
                    PinInAddScaXY.GetValue2D(i, out inSca.x, out inSca.y);
                    PinInAddVelXY.GetValue2D(i, out inVel.x, out inVel.y);
                    PinInAddIndex.GetValue(i, out inSlide);

                    _SlideList.Add(new SMTSlide(inPos, inRot, inSca, inVel, System.Convert.ToInt32(inSlide), 0, _SlideList.Count, UniqueSlideID++));
                }
            }
            
        }

        void fnReset()
        {
            _SlideList.Clear();
        }

        void fnFindCursors()
        {
            //go through all cursors
            //find out if they already exist
            //if yes, then update the coordinates, if no, make it, assign it to a slide
            //remove all dead cursors

            //tell all cursors that they might die this frame 
            for (int i = 0; i < _CursorList.Count; i++)
                _CursorList[i].youMightDie();

            //sort the slides so we know which one to look at first
            _SlideList.Sort();

            //read into list all fingers
            double currentx, currenty, currentid;
            Vector2D slidexy;
            for (int i = 0; i < Math.Min(PinInFingerXY.SliceCount, PinInFingerID.SliceCount); i++)
            {
                PinInFingerXY.GetValue2D(i, out currentx, out currenty);
                PinInFingerID.GetValue(i, out currentid);

                //if same ID as existing cursor, then update that cursor
                //otherwise make a new one
                if (_CursorList.Exists(new System.Predicate<SMTCursor>(delegate(SMTCursor thiscur) { return thiscur.isID(currentid); })))
                {
                    int idx_cursor = _CursorList.FindIndex(new System.Predicate<SMTCursor>(delegate(SMTCursor thiscur) { return thiscur.isID(currentid); }));
                    _CursorList[idx_cursor].isGoingToDie = false;
                    _CursorList[idx_cursor].xy = new Vector2D(currentx, currenty);
                }
                else // make a new cursor, set it up
                {
                    _CursorList.Add(new SMTCursor(currentx, currenty, System.Convert.ToInt32(currentid), UniqueCursorID++));

                    //does it hit a slide?
                    for (int iSlide = 0; iSlide < _SlideList.Count; iSlide++)
                    {
                        slidexy = (!_SlideList[iSlide].getMatrix() * new Vector2D(currentx, currenty)).xy;
                        List<SMTCursor> FoundCursorsHere = _CursorList.FindAll(new System.Predicate<SMTCursor>(delegate(SMTCursor thiscur) { return thiscur.slideindex == _SlideList[iSlide].UniqueID; }));
                        if (Math.Abs(slidexy.x) < 1 && Math.Abs(slidexy.y) < 1)
                        {
                            _CursorList[_CursorList.Count - 1].hitSlide = true;
                            _CursorList[_CursorList.Count - 1].slideindex = _SlideList[iSlide].UniqueID;
                            if (FoundCursorsHere.Count == 0)
                            {
                                _SlideList[iSlide].cursoruniqueid = _CursorList[_CursorList.Count - 1].uniqueid;
                                _SlideList[iSlide].isMoving = true;
                            }
                            else if (FoundCursorsHere.Count == 1)
                            {
                                _CursorList[_CursorList.Count - 1].startMT();
                                FoundCursorsHere[0].startMT();
 
                                _SlideList[iSlide].cursorstartMTxy1 = FoundCursorsHere[0].xy;
                                _SlideList[iSlide].cursorstartMTxy2 = new Vector2D(currentx, currenty) ;
                                _SlideList[iSlide].isMT = true;
                            }

                            break;
                        }
                    }
                }

            }

            //kill all cursors that are dead
            int iFoundSlide;

            _CursorList.RemoveAll(new System.Predicate<SMTCursor>(delegate(SMTCursor thiscur) {
                if (thiscur.isGoingToDie)
                {
                    //Find any slides that are connected to this cursor
                    iFoundSlide = _SlideList.FindIndex(new System.Predicate<SMTSlide>(delegate(SMTSlide thisslide) { return thisslide.UniqueID==thiscur.slideindex; }));
                    if (iFoundSlide != -1)
                    {
                        if (_SlideList[iFoundSlide].cursoruniqueid.Equals(thiscur.uniqueid))
                            _SlideList[iFoundSlide].isMoving = false;
                        else
                        {
                            _SlideList[iFoundSlide].isMT = false;
                        }
                    }
                }
                return thiscur.isGoingToDie;
            }
            ));
        }

        /* old fnFindIntersects
        void fnFindIntersects()
        {
            // find all intersects

            _SlideList.Sort();

            // calc slide cursors
            _IntersectList.Clear();
            Vector2D currentxy;

            //tell all cursors that they might die this frame 
            for (int i = 0; i < _CursorList.Count; i++)
                _CursorList[i].hitSlide=false;

            for (int iSlide = 0; iSlide < _SlideList.Count; iSlide++)
                _SlideList[iSlide].clearCursor();

            for (int iCursor = 0; iCursor < _CursorList.Count; iCursor++)
            {
                for (int iSlide = 0; iSlide < _SlideList.Count; iSlide++)
                {

                    currentxy = (!_SlideList[iSlide].getMatrix() * _CursorList[iCursor].xy).xy;
                    if (Math.Abs(currentxy.x)<1 && Math.Abs(currentxy.y)<1)
                    {

                        _IntersectList.Add(new SMTIntersect(System.Convert.ToInt32(_CursorList[iCursor].index),
                            iSlide,
                            _CursorList[iCursor].xy,
                            currentxy));
                        if (!_SlideList[iSlide].hasCursor())
                        {
                            _SlideList[iSlide].setCursor(System.Convert.ToInt32(_CursorList[iCursor].index), _CursorList[iCursor].xy);
                            _SlideList[iSlide].doTouch();
                        }
                        _CursorList[iCursor].hitSlide = true && _CursorList[iCursor].hitSlide;
                        break;

                    }
                }
            }

            //kill all cursors that are dead
            _CursorList.RemoveAll(new System.Predicate<SMTCursor>(delegate(SMTCursor thiscur) { return !thiscur.hitSlide; }));
            

        }
        */

        void fnFindIntersects()
        {
            Vector3D SlideCursor;

            _IntersectList.Clear();

            for (int i = 0; i < _SlideList.Count; i++)
                for (int j = 0; j < _CursorList.Count; j++)
                {
                    if (_CursorList[j].slideindex == _SlideList[i].UniqueID)
                    {
                        SlideCursor = new Vector3D(_CursorList[j].xy.x, _CursorList[j].xy.y, 0);
                        SlideCursor = !_SlideList[i].getMatrix() * SlideCursor;
                        _IntersectList.Add(new SMTIntersect(_CursorList[j].uniqueid, _SlideList[i].UniqueID, _CursorList[j].xy, SlideCursor.xy));
                    }
                }
        }

        void fnMaybeClose()
        {
            double closewidth, closeheight;
            PinInSizeCloseButton.GetValue2D(0,out closewidth, out closeheight);

            for (int i = 0; i < _IntersectList.Count; i++)
            {
                if (_IntersectList[i].getSlideXY().x>1-(2*closewidth)
                    && _IntersectList[i].getSlideXY().y>1-(2*closeheight))
                    _SlideList.RemoveAll(
                        new System.Predicate<SMTSlide>(delegate(SMTSlide thisslide) {
                            return thisslide.UniqueID == _IntersectList[i].SlideID;
                        }));
            }
        }

        void fnMoveSlides()
        {
            //go through slides
            //does a cursor exist for this slide?
            //if 1 cursor, then move
            //if 2 cursors then let's jeff han!

            int count = _SlideList.Count;
            List<SMTCursor> foundCursors;

            _CursorList.Sort();

            for (int i = 0; i < count; i++)
            {
                //search for cursor that hits this slide
                //use that cursor's change in xy since creation as a translate
                foundCursors = _CursorList.FindAll(new System.Predicate<SMTCursor>(delegate(SMTCursor thiscur) { return thiscur.slideindex==_SlideList[i].UniqueID; }));
                if (foundCursors.Count == 1)
                {
                    _SlideList[i].setdeltaxy(foundCursors[0].deltaxy);
                } else if(foundCursors.Count>1)
                {

                    #region Jeff Han
                    _SlideList[i].setdeltaxy(foundCursors[0].deltaxy);
                    _SlideList[i].primarycursorxy = foundCursors[0].xy;

                    //scale difference
                    double scalefactor = VMath.Dist(foundCursors[0].xy, foundCursors[1].xy) /
                        VMath.Dist(_SlideList[i].cursorstartMTxy1, _SlideList[i].cursorstartMTxy2);
                    _SlideList[i].setdeltascale(scalefactor);

                    double rottheta = -Math.Atan2(foundCursors[1].xy.x - foundCursors[0].xy.x, foundCursors[1].xy.y - foundCursors[0].xy.y) -
                        -Math.Atan2(foundCursors[1].startMTxy.x - foundCursors[0].startMTxy.x, foundCursors[1].startMTxy.y - foundCursors[0].startMTxy.y);
                    _SlideList[i].setdeltarot(rottheta);
                    #endregion

                }

                //if(_CursorList.Exists(new System.Predicate<SMTCursor>(delegate(SMTCursor thiscur) { return thiscur.slideindex==_SlideList[i].UniqueID; })))
                //    _SlideList[i].isMoving=true;

            }
            // move 
        }

        void fnOutput()
        {
            int count = _SlideList.Count;
            PinOutIndex.SliceCount = count;
            PinOutTransform.SliceCount = count;
            PinOutOver.SliceCount = count;

            PinOutCount.SetValue(0, count);

            _SlideList.Sort();

            for (int i = 0; i < count; i++)
            {
                PinOutIndex.SetValue(i, _SlideList[count-i-1].LastCursorID);
                PinOutTransform.SetMatrix(i, _SlideList[count-i-1].getMatrix());
                PinOutOver.SetValue(i,(_SlideList[count-i-1].isMoving ? 1 : 0));
            }


            PinOutSlideCursor.SliceCount = _IntersectList.Count;

            for (int i = 0; i < _IntersectList.Count; i++)
                PinOutSlideCursor.SetValue2D(i, _IntersectList[i].getSlideXY().x, _IntersectList[i].getSlideXY().y);
        }
        #endregion mainloop
    }
}
